/*
 *
 * Component:
 *
 * Description:
 *
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef _TASKD_JOBMAN_H_INCLUDED_
#define _TASKD_JOBMAN_H_INCLUDED_
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

#include <string>
#include <map>

using namespace std;

class JobManager;

/**
 * Class holding job and job file information
 */
class JobInfo {
	friend class JobManager;
	private:

		/** Job ID */
		string m_id;

		/** Initiating user */
		string m_user;

		/** Job Name */
		string m_name;

		/** Job Information formatted for creatTaskFile */
		string m_jobinfo;
		
		/** Name of the file storing job information */
		string m_fileName;

		/** Time the job information was last touched */
		time_t m_lastTouched;
	
		/**
		 * Disabled Copy Constructor
		 */
		JobInfo( const JobInfo & orig ) {
			m_fileName = orig.m_fileName;
			m_id = orig.m_id;
			m_user = orig.m_user;
			m_name = orig.m_name;
			m_lastTouched = orig.m_lastTouched;
		}

		/**
		 * Disabled assignment operator
		 */
		JobInfo & operator=( const JobInfo & orig ) {
			m_fileName = orig.m_fileName;
			m_id = orig.m_id;
			m_user = orig.m_user;
			m_name = orig.m_name;
			m_lastTouched = orig.m_lastTouched;

			return *this;
		}

	protected:
		
		/**
		 * Constructor for creating a new Job Info object.
		 *
		 * @param fileName   Name of the file holding the Job Information
		 * @param id         Job ID
		 * @param user       Initiating User
		 * @param name       Job Name
		 */
		JobInfo( const string & fileName,  const string & id, const string & user, const string & name );

		/**
		 * @return the time the job (or job file, if any) was last touched.
		 */
		inline time_t getLastTouched() {
			struct stat fs;
			if ( 0 == stat( m_fileName.c_str(), &fs ) ) {
				m_lastTouched = fs.st_mtime;
			}
			return m_lastTouched;
		}

		/**
		 * Touches the job (or job file) if any.
		 */
		inline void touch() {
			utime( m_fileName.c_str(), NULL );
			m_lastTouched = time(NULL);
		}

		/**
		 * Creates a job file from Job Information if job file not yet exists.
		 */
		void createFile();
		
		/**
		 * @return Name of the file holding job information
		 */
		inline const char * getJobFile() const {
			return m_fileName.empty() ? NULL : m_fileName.c_str();
		}
		
	public:

		/**
		 * @return Job ID or NULL
		 */
		inline const char * getId() const {
			return m_id.empty() ? NULL : m_id.c_str();
		}

		/**
		 * @return Initiating user or NULL
		 */
		inline const char * getUser() const {
			return m_user.empty() ? NULL : m_user.c_str();
		}

		/**
		 * @return Job name or NULL
		 */
		inline const char * getName() const {
			return m_name.empty() ? NULL : m_name.c_str();
		}

		/**
		 * @return Job information formatted for createTaskFile or NULL
		 */
		inline const char * getInfo() const {
			return m_jobinfo.empty() ? NULL : m_jobinfo.c_str();
		}
		
		/**
		 * Checks if the Job Information (including job ID) is the same as in the JobInfo object argument.
		 *
		 * @param ji   JobInfo object this object be compared to
		 *
		 * @return true if and only if the job ID, Initiating User and Job name is the same in the two JobInfo objects
		 */
		inline bool equals( JobInfo * ji ) {
			return m_id.compare( ji->m_id ) == 0 && m_user.compare( ji->m_user ) == 0 && m_name.compare( ji->m_name ) == 0;
		}
};


/**
 * Managing Job IDs in an application where certain tasks need to be organized into the same job.
 * 
 * Job IDs are associated to Task IDs, that is a Task ID connects the tasks to be organized to a Job.
 * The application need to generate the Task ID to identify the tasks be groupped to the same job.
 * JobManager will look up or generate Job ID/Job Information for the Task ID generated by the application.
 *
 * Generated Job Information will be stored in files in a Job Info directory. The file name is the same as the Task ID.
 * Caller should be aware the Job Info directory exists.
 *
 * Job Information passed by other applications should be placed to the same Job Info directory, and will be picked up by JobManager.
 *
 * Job Informations may expire after given timeout. If timeout expires, a new Job ID will be generated for the Task ID.
 * Timing restarts when each time the Job ID was requested.
 *
 * The number of maximum Job IDs tracked can be set. If the number of tracked Job IDs is reached, expired Job IDs will be dropped.
 * If there is no expired Job ID, the oldest Job IDs (the one having last requested the earliest) will be dropped.
 */
class JobManager {
	private:
		/** Maps task IDs to job IDs */
		map< string, JobInfo * > m_jobs;

		/** Iterator type for jobs */
		typedef map< string, JobInfo * >::iterator JobIterator;

		/** Directory to get local (or synchronized) job IDs for taskd IDs */
		string m_dirname;

		/** Expires job IDs after the given timeout (in seconds) */
		time_t m_expireTime;

		/** Maximum number of jobs to store */
		size_t m_maxNum;
		
		/** Default Job Name - if Job Info contains no Job Name */
		string m_defaultName;
		
		/** Default Initiating User - if Job Info contains no Initiating User */
		string m_defaultUser;
		
		/**
		 * Reads Job Information from a job file for a task
		 *
		 * @param jobFile  Path to the job file to read
		 * @param taskId   Identifies the tasks the job info be returned to
		 *
		 * @return the job info read for the task or NULL if an error occurs
		 */
		JobInfo * readJobFile( const string & jobFile, const string & taskId );

		/**
		 * Creates Job Information for a task
		 *
		 * @param jobFile  Path to the job file to read
		 * @param taskId   Identifying the tasks the job info be created for
		 *
		 * @return the job info created for the task
		 */
		JobInfo * createJob( const string & jobFile, const string & taskid );
			
	public:

		/**
		 * Constructor for creating a new JobManager object.
		 *
		 * @param dirname       Path to job info directory (job files be placed/searched for)
		 * @param maxNum        Maximum number of jobs to store
		 * @param expireTime    Time in seconds to expire Job IDs
		 * @param defaultUser   Default initiating user be set when generating new job info
		 * @param defaultName   Default job name be set when generating new job info
		 */
		JobManager( const string & dirname, size_t maxNum, time_t expireTime, const string & defaultUser, const string & defaultName );

		/**
		 * Constructor for creating a new JobManager object. Directory name should be set using setDir()
		 *
		 * @param maxNum        Maximum number of jobs to store
		 * @param expireTime    Time in seconds to expire Job IDs
		 * @param defaultUser   Default initiating user be set when generating new job info
		 * @param defaultName   Default job name be set when generating new job info
		 */
		JobManager( size_t maxNum, time_t expireTime, const string & defaultUser, const string & defaultName );

		/**
		 * Destructor for destroying the Job Manager
		 */
		~JobManager();

		/**
		 * Looks up Job Info for given task. If no Job Info was found, a new one will be generated.
		 *
		 * @param taskId   Identifying the tasks the job info be searched for
		 *
		 * @return Job Info found or generated for the task, or NULL if an error occured
		 */
		JobInfo * findJobInfo( const string & taskId );

		/**
		 * @return the number of jobids currently in the memory
		 */
		inline size_t size() const {
			return m_jobs.size();
		}

		/**
		 * @return path to job info directory
		 */
		inline const char * getDir() const {
			return m_dirname.empty() ? NULL : m_dirname.c_str();
		}

		/**
		 * Sets directory name for storing job info files.
		 * Directory name will only be set if not yet set before.
		 *
		 * @param dirname   Path to job info directory (job files be placed/searched for)
		 */
		inline void setDir( const string & dirname ) {
			if ( m_dirname.empty() )
				m_dirname = dirname;
		}
};

#endif // _TASKD_JOBMAN_H_INCLUDED_
