/*
 * Component: taskd
 *
 * Description:
 *     Header file for scheduling and executing tasks
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef SCHEDULER_H_INCLUDED_
#define SCHEDULER_H_INCLUDED_
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "taskd.h"
#include "queue.h"
#include "mymutex.h"
#include "mythread.h"
#include "fileparser.h"
#include "task.h"
#include "javacall.h"

/**
 * For comapring tasks in the retry queue
 */
class RetryComparator {
	public:
		inline bool operator()( const CTask * t1, const CTask * t2 ) const {
			return t1->EarlierThan( t2 );
		}
};

/**
 * Schedules tasks to be performed.
 */
class CScheduler {
	private:
		/** For iterating tasks in the scheduled queue */
		typedef set< CTask * >::iterator POSITION;

		/** For iterating tasks in the retry queue */
		typedef set< CTask *, RetryComparator >::iterator RETRY_POSITION;

	protected:
		/** The scheduled queue. Tasks to be executed will be put to this queue in execution order */
		Queue< CTask > m_lsTaskQueue;

		/** The retry queue. Tasks to be retried will be put to this queue in retry order */
		Queue< CTask, RetryComparator > m_lsRetryQueue;

		/** Array of running tasks */
		CTask ** m_pRunArray;

		/** Array of connections to the Java task manager */
		JavaCall ** m_pJavaArray;

		/** Path to the queue directory */
		char * m_strQueueDirectory;

		/** Length of m_strQueueDirectory */
		int m_nQueueDirNameLength;

		/** Host name or IP address of the Java task manager */
		char * m_strJavaServerIP;

		/** Port number of the Java task manager */
		int m_nJavaServerPort;

#ifndef CHECK_INTEGRITY
		MyMutex m_pSchedulerMutex;
#endif

		/** Maximum number of (concurrent) threads */
		int m_nMaxNumOfThreads;

		/** Number of threads to be used for normal priority tasks */
		int m_nNumOfNormalThreads;

		/** High priority boundary. Tasks with a higher priority than m_nPriorityBound are handled as high priority tasks.
		 *  (a smaller value means a higher priority) */
		int m_nPriorityBound;

		/** Counter of currently running threads */
		int m_nThreadCounter;

		/** Counter of the tasks in the scheduled queue */
		int m_nTaskCounter;

	private:

		/**
		 * Add a task to the retry queue.
		 *
		 * @param pTask   task to be added to the retry queue
		 */
		void AddToRetryQueue( CTask * pTask );

		/**
		 * Add a task to the scheduled queue.
		 *
		 * @param pTask   task to be added to the scheduled queue
		 */
		void AddToQueue( CTask * pTask );

		/**
		 * Initialize a queue.
		 *
		 * @param dt type of the queue
		 */
		void InitQueue( DirType dt );

		/**
		 * Starts the first task(s) from the scheduled queue (if any) in a new thread.
		 * Tasks will only be started if the number of running threads is within the limit.
		 */
		void StartNextTask();

		/**
		 * Checks if a task is running.
		 *
		 * @param pTask   task to be checked
		 *
		 * @return true if and only if the task is currently running
		 */
		bool IsRunning( CTask * pTask );

		/**
		 * Checks if there is an empty slot for starting new tasks and return the index of the first slot found.
		 *
		 * It will also count and return
		 *    - the total number of empty slots,
		 *    - the number of normal priority tasks running
		 *    - the highest priority of running tasks
		 *
		 * @param nEmptySlotCnt[out]        count of empty slots
		 * @param nNormalThreadCnt[out]     count of normal threads running
		 * @param nHighestPriority[out]     highest priority of running tasks
		 * @param pTask[in]                 the task just finished when called from TaskFinished [optional]
		 * @param pbDelete[out]             will be set to true if the task to be deleted [optional]
		 *
		 * @return the index of the first empty slot or -1 if there is no empty slot
		 */
		int getEmptySlotIndex( int & nEmptySlotCnt, int & nNormalThreadCnt, int & nHighestPriority, CTask * pTask = NULL,
				bool * pbDelete = NULL );

		/**
		 * Checks if a task is startable or not.
		 *
		 * A task is startable if the task is a high priority task or the number of concurrent normal tasks within the limit.
		 * The only exception is when the currently running highest priority task has significantly
		 * higher priority than the one to be started.
		 *
		 * @param pTask                task to be started
		 * @param nNormalThreadCnt     count of normal threads (ie. count of normal priority tasks running)
		 * @param nHighestPriority     highest priority of running tasks
		 *
		 * @return true if and only if the task is startable
		 */
		bool isTaskStartable( CTask * pTask, int nNormalThreadCnt, int nHighestPriority );

	public:
#ifdef CHECK_INTEGRITY
		MyMutex m_pSchedulerMutex;
#endif

		/**
		 * Create a new scheduler object.
		 *
		 * @param strDirName           path to the directory where the tasks are stored
		 * @param nMaxNumOfThreads     maximum number of threads (default is 8)
		 * @param nNumOfNormalThreads  maximum number of normal (non high priority) threads (default is 4)
		 * @param nPriorityBound       high priority boundary (default is 20)
		 */
		CScheduler( char * strDirName, int nMaxNumOfThreads = 8, int nNumOfNormalThreads = 4, int nPriorityBound = PRIORITY_BOUND_HIGH );

		/**
		 * Destructor for destroying the scheduler.
		 */
		~CScheduler();

		//	void SetTaskPriority(CTask * pTask, int nTaskPriority);

		/**
		 * Check if there are any new tasks in the incoming queue, and move them to
		 * the scheduled queue.
		 */
		void CheckAndAddNewTasks();

		/**
		 * Move a task from the scheduled queue to the retry queue.
		 *
		 * @param pTask  the task to be moved from the scheduled queue to the retry queue
		 */
		void MoveToRetryQueue( CTask * pTask );

		/**
		 * Reschedule retry tasks reached their retry scheduled time.
		 * Checks if there are retry tasks reached their retry scheduled time and
		 * move them from the retry queue to the scheduled queue.
		 */
		void CheckAndGetRetryTasks();

		/**
		 * Returns the scheduler mutex for locking.
		 *
		 * @return scheduler mutex can be used for locking
		 */
		inline MyMutex GetMutex() const {
			return m_pSchedulerMutex;
		}

		/**
		 * Adds the task to the scheduled queue (from the incoming queue)
		 *
		 * @param pTask   task to be added to the scheduled queue
		 */
		void AddTask( CTask * pTask );

		/**
		 * Removes the finished task from the scheduled queue and picks up the next task
		 * to be executed in the current thread.
		 * The finished task will be removed from the scheduled queue and will be:
		 *  - deleted if nRetrySec is 0 (=OK)
		 *  - moved to the retry queue if nRetrySec is greater than 0 (=ERROR)
		 *  - moved to the failed queue if nRetrySec is less than 0 (=FAILED)
		 *
		 * The next task will be picked up from the scheduled queue (if any) and will be started in the current thread.
		 *
		 * @param pTask      the task just finished
		 * @param pThread    the current thread
		 * @param nRetrySec  return value of the task
		 *
		 * @return the next task started in the current thread (might be NULL)
		 */
		CTask * TaskFinished( CTask * pTask, MyThread * pThread, int nRetrySec );

		/**
		 * Returns the number of the tasks in the scheduled queue.
		 *
		 * @return the number of the tasks in the scheduled queue
		 */
		inline int GetTaskCounter() {
			return m_nTaskCounter;
		}

		/**
		 * Returns the full path to the queue directory.
		 *
		 * @return full path to the queue directory
		 */
		inline char * GetQueueDirectory() {
			return m_strQueueDirectory;
		}

		/**
		 * Creates full path for a task to a queue.
		 * The returned value should be freed.
		 *
		 * @param strFileName   short file name of the task
		 * @param dt            queue type (scheduled, failed, retry, ...)
		 *
		 * @return full path to the task in the queue (should be freed)
		 */
		char * GetFullFileName( const char * strFileName, DirType dt ); // result should bee freed up.

		/**
		 * Creates full path for a task to a queue.
		 * The returned value should be freed.
		 *
		 * @param fp    file parser object of the task
		 * @param dt    queue type (scheduled, failed, retry, ...)
		 *
		 * @return full path to the task in the queue (should be freed)
		 */
		char *GetFullFileName( FileParser * fp, DirType dt );

		/**
		 * Sets the Java task manager properties for running java tasks
		 *
		 * @param strIP    IP address (host name) of the Java task manager
		 * @param nPort    Port number of the Java task manager
		 */
		void SetJavaServer( char * strIP, int nPort );

		/**
		 * Returns a java connection of the given slot.
		 * (Slot is an index in the array of java connections).
		 * If there is no java connection on the given slot,
		 * a new one will be created and associated to the given slot.
		 *
		 * @param nSlot return java connection of this slot
		 *
		 * @return the java connection
		 */
		JavaCall * GetJavaCall( int nSlot );

		/**
		 * Waits for all threads to be finished.
		 */
		void waitForThreads();

		/**
		 * Waits for the threads to be finished within timeOut.
		 * The timeOut parameter will be updated (decremented), so
		 * the caller will know the remaining timeout.
		 * No infinite timeOut can be specified
		 *
		 * @param timeOut   time out in milliseconds
		 *
		 * @return true if and only if all threads finished within the timeout
		 */
		bool waitForThreads( unsigned int & timeOut );

		// Check Integrity
#ifdef CHECK_INTEGRITY
		bool IsInRunArray( CTask * t );
		void CheckIntegrity( const char * msg, CTask * pTask, bool bShouldBeInQueue = true, CTask * timeTask = NULL );

		void CheckFile( const char * msg, CTask * pTask, DirType dt ) {
			char *fullname = GetFullFileName( pTask->GetFileParser(), dt );
			struct stat st;
			int resstat = stat( fullname, &st );

			if ( resstat != 0 ) {
#ifdef TRACE
				trace("SIGTERM raised from CheckFile: %s::%p %s [%d, %d]", msg, pTask, fullname, resstat, errno);
#endif
				raise( SIGTERM);
			}
			delete fullname;
		}
#endif
};

#endif // SCHEDULER_H_INCLUDED_
