/*
 * Component: taskd
 *
 * Description:
 *     Header file for shell, java and so task implementations
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef _TASK_H_INCLUDED___
#define _TASK_H_INCLUDED___

#include "mythread.h"
#include "fileparser.h"
#include "taskd.h"

class CScheduler;

/**
 * Abstract base class of tasks in taskd
 */
class CTask {
	// CScheduler can see protected methods
	friend class CScheduler;

 private:
	/** Task file name parser */
	FileParser * m_FileParser;

 protected:

	/** Thread in which the task will be run */
	MyThread *m_pThread;

	/** Is the task running */
	bool m_bRun;

	/** The sceduler object */
	CScheduler *m_pScheduler;

	/** Index of the task in the schedulers running tasks array */
	int m_nSlot;

	/** Assign the scheduler for this task. The task will be scheduled and started by
	 * the assigned scheduler.
	 *
	 * @param pScheduler   scheduler to be assigned to this task
	 */
	inline void SetScheduler(CScheduler * pScheduler) {
		m_pScheduler = pScheduler;
	}
	
 public:
	/**
	 * Constructor for creating a new task object.
	 *
	 * @param fp  task file in a parsed form
	 */ 
	CTask(FileParser * fp);

	/**
	 * Destructor for destroying a task object.
	 */
	virtual ~ CTask();

	/**
	 * Main method of the task to be performed.
	 * Derived classes should override this method.
	 *
	 * @return 0  if performing the task succeeded
	 * @return 0< if a temporary error occured and the task should be retried (= the seconds to retry)
	 * @return <0 if a permanent error occured and the task failed
	 */
	virtual int Run() = 0;

	/**
	 * Creates a file name to store the task content in. Used for moving the
	 * task from one queue to another.
	 * 
	 * The return value should be freed.
	 *
	 * @return the filename of the task or NULL
	 */
	inline char * CreateFileName() {
		return NULL == m_FileParser ? NULL : m_FileParser->CreateFileName();
	}

	/**
	 * Returns the number of retries of this task.
	 * (How many times the task tried to be performed)
	 *
	 * @return the number of retries
	 */
	inline int GetRetryNum() {
		return (NULL == m_FileParser) ? 0 : m_FileParser->GetRetryNum();
	}

	/**
	 * Increments the tasks retry number (by 1) and sets the seconds to the next retry.
	 *
	 * @param seconds   retry to perform the task after <code>seconds</code> seconds ellapsed.
	 */
	inline void IncrementRetryNum(int seconds) {
		if (NULL != m_FileParser) {
			m_FileParser->IncrementRetryNum();
			m_FileParser->SetDateTime(time(NULL) + seconds);
		}
	}

	/**
	 * Returns the tasks priority.
	 *
	 * @return task priority.
	 */
	inline int GetTaskPriority() {
		return NULL == m_FileParser ? PRIORITY_LOWEST : m_FileParser->GetPriority();
	}

	/**
	 * Returns the file parser.
	 * (the task file name in a parsed form)
	 *
	 * @return the file parser of this task
	 */
	inline FileParser *GetFileParser() const {
		return m_FileParser;
	}

	/**
	 * Returns the thread in which the task is being executed.
	 *
	 * @return the thread in which the task is being executed
	 */
	inline MyThread *GetThread() {
		return m_pThread;
	}

	/**
	 * Assigns a thread to this task. The task will be executed in this thread.
	 * Also sets the 'Run' flag.
	 *
	 * @param thread   thread to be assigned to this task
	 */
	inline void SetThread(MyThread * pThread) {
		m_pThread = pThread;
		m_bRun = true;
	}

	/**
	 * Sets the index of the task in the running tasks array of the scheduler.
	 * Will be used by the scheduler when executing the task is being started.
	 *
	 * @param nSlot   the index of the task in the running tasks array of the scheduler
	 */
	inline void SetSlot(int nSlot) {
		m_nSlot = nSlot;
	}

	/**
	 * Checks if the task is running (being executed).
	 *
	 * @return true if and only if the task is running
	 */
	inline bool IsRun() {
		return m_bRun;
	}

	/**
	 * Returns the scheduler assigned to this task.
	 *
	 * @return the scheduler assigned to this task
	 */
	inline CScheduler * GetScheduler() {
		return m_pScheduler;
	}

	/**
	 * Compares this task to another task and checks if this task is less than the other one.
	 * task A is less then task B if the file parser object of A is less then the file parser object of B.
	 *
	 * This operator will be used to order the tasks in the scheduled queue.
	 *
	 * @see FileParser::LessThan
	 *
	 * @param t   task to be compared
	 *
	 * @return true if and only if this task is less than <code>t</code>
	 */
	inline bool operator<(const CTask & t) const {
		return m_FileParser->LessThan(t.GetFileParser());
	}
	
	/**
	 * Compares this task to another task and checks if this task is earlier than the other one.
	 * task A is earlier then task B if the file parser object of A is earlier then the file parser object of B.
	 *
	 * Will be used to order the tasks in the retry queue
	 *
	 * @see FileParser::EarlierThan
	 *
	 * @param t   task to be compared
	 *
	 * @return true if and only if this task is earlier than <code>t</code>
	 */
	inline bool EarlierThan(const CTask * t) const {
		return m_FileParser->EarlierThan(t->GetFileParser());
	}
#ifdef CHECK_INTEGRITY
 //For CheckIntegrity public:
	 bool m_bExists;
#endif
};


/**
 * For executing shell tasks.
 */
class ShellTask : public CTask {
 public:

	/**
	 * Constructor for creating a new shell task object.
	 * 
	 * @param fp  task file in a parsed form
	 */
	ShellTask(FileParser * fp) : CTask(fp) {
	}
	
	/**
	 * Destructor for destroying the shell task object.
	 */
	~ShellTask() {
	}

	/**
	 * Main method of the shell task to be performed.
	 *
	 * @return 0  if performing the task succeeded
	 * @return 0< if a temporary error occured and the task should be retried (= the seconds to retry)
	 * @return <0 if a permanent error occured and the task failed
	 */
	int Run();
};


/**
 * For executing java tasks
 */
class JavaTask : public CTask {
 public:

	/**
	 * Constructor for creating a new java task object.
	 * 
	 * @param fp  task file in a parsed form
	 */
	JavaTask(FileParser * fp) : CTask(fp) {
	}
	
	/**
	 * Destructor for destroying the java task object.
	 */
	~JavaTask() {
	}

	/**
	 * Main method of the java task to be performed.
	 *
	 * @return 0  if performing the task succeeded
	 * @return 0< if a temporary error occured and the task should be retried (= the seconds to retry)
	 * @return <0 if a permanent error occured and the task failed
	 */
	int Run();
};


/**
 * For executing so (C/C++) tasks
 */
class SoTask : public CTask {
 public:

	/**
	 * Constructor for creating a new so task object.
	 * 
	 * @param fp  task file in a parsed form
	 */
	SoTask(FileParser * fp):CTask(fp) {
	}
	
	/**
	 * Destructor for destroying the so task object.
	 */
	~SoTask() {
	}

	/**
	 * Main method of the so task to be performed.
	 *
	 * @return 0  if performing the task succeeded
	 * @return 0< if a temporary error occured and the task should be retried (= the seconds to retry)
	 * @return <0 if a permanent error occured and the task failed
	 */
	int Run();
};

#endif  // _TASK_H_INCLUDED___
